/********************************************************
 *
 *  Copyright (c) 2015 Sony Corporation
 *
 ********************************************************/

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/pci.h>
#include <linux/cdev.h>
#include <linux/time.h>
#include <linux/sched.h>
#include <linux/reboot.h>
#include <linux/vmalloc.h>
#include <linux/ioctl.h>
#include <linux/mm.h>
#include <linux/poll.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/interrupt.h>

#include <linux/errno.h>

#include <linux/of_platform.h>

#define KDEBUG
#include "exttype.h"
#include "general_macro.h"
#include "dbg_assert.h"

#include "elbdriver.h"

MODULE_LICENSE("Dual BSD/GPL");

DECLARE_WAIT_QUEUE_HEAD(vd_wait_queue);

typedef struct {
    void*   addr;
    ulong   size;
} LbmData_t;

static LbmData_t elb_cs2_data;
static LbmData_t elb_cs3_data;

static irqreturn_t elb_intr(int irq, void* data)
{
    wake_up_interruptible(&vd_wait_queue);

    return IRQ_HANDLED;
}

static int elb_wait_vd(void)
{
#define VD_TIMEOUT (5*1000)
    int ret_val = interruptible_sleep_on_timeout(&vd_wait_queue, VD_TIMEOUT);

    static ushort pre_vcounter = 0;
    static uchar  pre_oe = 0;

    uchar* addr     = elb_cs3_data.addr;
    uchar data1     = ioread8(addr + 0x8a);
    uchar data2     = ioread8(addr + 0x8b);
    ushort vcounter = UC2_TO_USHORT(data1, data2);
    uchar oe        = ioread8(addr+0x13);

#if 0 //for debug
    if (vcounter != 0) {
        if ((oe == pre_oe) || (vcounter != (pre_vcounter+1))) {
            printk("%s %d: %04x(%04x), %2x (%2x)\n", __func__, __LINE__,
                                                     vcounter, pre_vcounter, pre_oe, oe);
        }
    }
#endif
    if ((data1 == 0xff) && (data2 == 0xff)) {
        uchar data = 1;
        iowrite8(data, addr+0x87);

        udelay(1*100);

        data = 0;
        iowrite8(data, addr+0x87);
    }
    pre_vcounter = vcounter;
    pre_oe       = oe;

    if (ret_val < 0) {
        printk("%s: fault (%d). \n", __func__, ret_val);
        return ret_val;
    }

    return OK;
}

static int elb_ioctl_wait_vd(void)
{
    return elb_wait_vd();
}

static int lbm_ioctl_read8(void* base_addr, LbmdIoctl_t* ioctl_data)
{
    uchar* addr = base_addr + ioctl_data->offset;

    int     i;
    int     ret_val;
    uchar   data;
    uchar*  ptr = ioctl_data->buf;

    for (i = 0; i < ioctl_data->size / sizeof(data); ++i) {
        data = ioread8(addr);
        ++ addr;
        ret_val = copy_to_user(&ptr[i], &data, sizeof(data));

        if (ret_val) {
            printk("Error : %s : copy_to_user failed [%d]\n", __func__, ret_val);
        }
    }

    return ret_val;
}

static int lbm_ioctl_read16(void* base_addr, LbmdIoctl_t* ioctl_data)
{
    ushort* addr = base_addr + ioctl_data->offset;

    int     i;
    int     ret_val;
    ushort  data;
    ushort* ptr = ioctl_data->buf;

    for (i = 0; i < ioctl_data->size / sizeof(data); ++i) {
        data = ioread16be(addr);
        ++ addr;
        ret_val = copy_to_user(&ptr[i], &data, sizeof(data));

        if (ret_val) {
            printk("Error : %s : copy_to_user failed [%d]\n", __func__, ret_val);
        }
    }

    return ret_val;
}

static int lbm_ioctl_write8(void* base_addr, LbmdIoctl_t* ioctl_data)
{
    uchar* addr = base_addr + ioctl_data->offset;

    int     i;
    int     ret_val;
    uchar   data;
    uchar*  ptr = ioctl_data->buf;

    for (i = 0; i < ioctl_data->size / sizeof(data); ++i) {
        ret_val = copy_from_user(&data, &ptr[i], sizeof(data));
        iowrite8(data, addr);
        //kdebug("%s %s %d : %08x = %x\n", __FILE__, __FUNCTION__, __LINE__, addr, data);
        ++addr;

        if (ret_val) {
            printk("Error : %s : copy_from_user failed [%d]\n", __func__, ret_val);
        }
    }

    return ret_val;
}

static int lbm_ioctl_write16(void* base_addr, LbmdIoctl_t* ioctl_data)
{
    ushort* addr = base_addr + ioctl_data->offset;

    int     i;
    int     ret_val;
    ushort  data;
    ushort* ptr = ioctl_data->buf;

    for (i = 0; i < ioctl_data->size / sizeof(data); ++i) {
        ret_val = copy_from_user(&data, &ptr[i], sizeof(data));
        iowrite16be(data, addr);
        ++addr;

        if (ret_val) {
            printk("Error : %s : copy_from_user failed [%d]\n", __func__, ret_val);
        }
    }

    return ret_val;
}

int elb_fops_ioctl(struct inode* inode, struct file* filep, uint cmd, ulong arg)
{
    int             ret = OK;
    LbmdIoctl_t*    ioctl_data = (LbmdIoctl_t __user *)arg;
    void*           base_addr = ioctl_data->area == LBM_AREA_CS2 ?
                                    elb_cs2_data.addr : elb_cs3_data.addr;

    switch (cmd) {
    case IOCTL_READ:
        if (!access_ok(VERIFY_WRITE, (LbmdIoctl_t __user *)arg, _IOC_SIZE(cmd))) {
            kdebug("%s %s %d\n", __FILE__, __FUNCTION__, __LINE__);
            return -EFAULT;
        }
        if (ioctl_data->area == LBM_AREA_CS2) {
            ret = lbm_ioctl_read16(base_addr, ioctl_data);
        }
        else {
            ret = lbm_ioctl_read8(base_addr, ioctl_data);
        }
        break;

    case IOCTL_WRITE:
        if (!access_ok(VERIFY_READ, (LbmdIoctl_t __user *)arg, _IOC_SIZE(cmd))) {
            kdebug("%s %s %d\n", __FILE__, __FUNCTION__, __LINE__);
            return -EFAULT;
        }
        if (ioctl_data->area == LBM_AREA_CS2) {
            ret = lbm_ioctl_write16(base_addr, ioctl_data);
        }
        else {
            ret = lbm_ioctl_write8(base_addr, ioctl_data);
        }
        break;

    case IOCTL_WAIT_VD:
        ret = elb_ioctl_wait_vd();
        break;

    default:
        printk("%s: Command should be read or write.\n", __func__);
        return ERROR;
    }

    return ret;
}

static int elb_fops_open(struct inode* inode, struct file* file)
{
   //kdebug("%s: major/minor =  %d/%d (pid %d)\n",
   //       __func__, imajor(inode), iminor(inode), current->pid);

    inode->i_private = inode;
    file->private_data = file;

    return OK;
}

static int elb_fops_close(struct inode* inode, struct file* file)
{
   //kdebug("%s: major/minor =  %d/%d (pid %d)\n",
   //       __func__, imajor(inode), iminor(inode), current->pid);

    return OK;
}

struct file_operations elb_fops = {
    .owner   = THIS_MODULE,
    .open    = elb_fops_open,
    .release = elb_fops_close,
    .ioctl   = elb_fops_ioctl,
};

//------------------------------------------------------------------
static int          elb_devs  = 1;
static int          elb_major = 0;
static struct cdev  elb_cdev;
static int elb_io_init(void)
{
    dev_t dev = MKDEV(elb_major, 0);
    int ret_val  = 0;
    int major;

    ret_val = alloc_chrdev_region(&dev, 0, elb_devs, DEVICE_NAME);
    if (ret_val) {
        return ret_val;
    }

    cdev_init(&elb_cdev, &elb_fops);
    elb_cdev.owner = THIS_MODULE;
    elb_major = major = MAJOR(dev);

    ret_val = cdev_add(&elb_cdev, MKDEV(elb_major, 0), elb_devs);
    if (ret_val) {
        return ret_val;
    }

    elb_cs2_data.addr = ioremap_nocache(LBM_CS2_ADDRESS, LBM_CS2_SIZE);
    elb_cs2_data.size = LBM_CS2_SIZE;

    elb_cs3_data.addr = ioremap_nocache(LBM_CS3_ADDRESS, LBM_CS3_SIZE);
    elb_cs3_data.size = LBM_CS3_SIZE;

    if (elb_cs2_data.addr == NULL) {
        printk("%s: Allocate Error : Physical Addr[0x%08x], Size[0x%08x]\n",
                __func__, LBM_CS2_ADDRESS, LBM_CS2_SIZE);
        return -ENOMEM;
    }

    if (elb_cs3_data.addr == NULL) {
        printk("%s: Allocate Error : Physical Addr[0x%08x], Size[0x%08x]\n",
                __func__, LBM_CS3_ADDRESS, LBM_CS3_SIZE);
        return -ENOMEM;
    }

    return OK;
}

static int elb_io_exit(void)
{
    dev_t dev = MKDEV(elb_major, 0);

    cdev_del(&elb_cdev);
    unregister_chrdev_region(dev, elb_devs);

    iounmap(elb_cs2_data.addr);
    iounmap(elb_cs3_data.addr);

    return OK;
}

#define INTR_EXTERNAL_IRQ0 0
//------------------------------------------------------------------
static int elb_probe(struct of_device* ofdev, const struct of_device_id* match)
{
    struct device_node* np = ofdev->node;
    int                 virq = 256;
    int                 ret_val;

    if (!of_device_is_available(np)) {
        printk("%s %d: ERROR\n", __func__, __LINE__);
        return -ENODEV;
    }

    virq = of_irq_to_resource(np, INTR_EXTERNAL_IRQ0, NULL);
    if (virq == NO_IRQ) {
        printk("%s %d: ERROR\n", __func__, __LINE__);
        return -EINVAL;
    }

    printk("virq : %d\n", virq);

    ret_val = request_irq(virq,
                          &elb_intr,
                          0, //<--- tako
                          DEVICE_NAME,
                          NULL);
    if (ret_val < 0) {
        printk("%s: Failed to set interrupt handler(%d) %d\n", __func__, virq, ret_val);
        printk(" - code %d , EIO %d , EINVAL %d\n", ret_val, EIO, EINVAL);
        of_node_put(np);
        return ret_val;
    }

    ret_val = elb_io_init();

    printk("%s %s : %s driver installed. \n", __DATE__, __TIME__, DEVICE_NAME);

    return ret_val;
}

static int elb_remove(struct of_device* ofdev)
{
    struct device_node* np = ofdev->node;
    int                 virq = 256;

    if (!of_device_is_available(np)) {
        printk("%s %d: ERROR\n", __func__, __LINE__);
        return -ENODEV;
    }

    virq = of_irq_to_resource(np, INTR_EXTERNAL_IRQ0, NULL);
    if (virq == NO_IRQ) {
        printk("%s %d: ERROR\n", __func__, __LINE__);
        return -EINVAL;
    }

    printk("virq : %d\n", virq);

    free_irq(virq, NULL);

    elb_io_exit();

    printk("driver %s unloaded\n", DEVICE_NAME);

    return OK;
}

static struct of_device_id exintr_match[] __devinitdata = {
    { .compatible = "fsl,mpic-exintr", },
    { /* end of list */ },
};

MODULE_DEVICE_TABLE(of, exintr_match);
static struct of_platform_driver exintr_driver = {
    .name           = DEVICE_NAME,
    .owner          = THIS_MODULE,
    .match_table    = exintr_match,
    .probe          = elb_probe,
    .remove         = elb_remove,
};

static int __init elb_init(void)
{
    printk("driver %s \n", DEVICE_NAME);
    return of_register_platform_driver(&exintr_driver);
}

static void __exit elb_exit(void)
{
    printk("driver %s \n", DEVICE_NAME);
    of_unregister_platform_driver(&exintr_driver);
}

module_init(elb_init);
module_exit(elb_exit);
